from omlish.lite.marshal import ObjMarshalerManager
from omlish.lite.marshal import PolymorphicObjMarshaler

from .base import Command
from .base import CommandNameMap


##


def install_command_marshaling(
        cmds: CommandNameMap,
        msh: ObjMarshalerManager,
) -> None:
    for fn in [
        lambda c: c,
        lambda c: c.Output,
    ]:
        msh.set_obj_marshaler(
            fn(Command),
            PolymorphicObjMarshaler.of([
                PolymorphicObjMarshaler.Impl(
                    fn(cmd),
                    name,
                    msh.get_obj_marshaler(fn(cmd)),
                )
                for name, cmd in cmds.items()
            ]),
        )



# ruff: noqa: UP006 UP007 UP045
import abc
import dataclasses as dc
import logging
import traceback
import typing as ta

from omlish.lite.abstract import Abstract
from omlish.lite.check import check
from omlish.lite.strings import snake_case


CommandT = ta.TypeVar('CommandT', bound='Command')
CommandOutputT = ta.TypeVar('CommandOutputT', bound='Command.Output')


##


@dc.dataclass(frozen=True)
class Command(Abstract, ta.Generic[CommandOutputT]):
    @dc.dataclass(frozen=True)
    class Output(Abstract):
        pass

    @ta.final
    async def execute(self, executor: 'CommandExecutor') -> CommandOutputT:
        return check.isinstance(await executor.execute(self), self.Output)  # type: ignore[return-value]


##


@dc.dataclass(frozen=True)
class CommandException:
    name: str
    repr: str

    traceback: ta.Optional[str] = None

    exc: ta.Optional[ta.Any] = None  # Exception

    cmd: ta.Optional[Command] = None

    @classmethod
    def of(
            cls,
            exc: Exception,
            *,
            omit_exc_object: bool = False,

            cmd: ta.Optional[Command] = None,
    ) -> 'CommandException':
        return CommandException(
            name=type(exc).__qualname__,
            repr=repr(exc),

            traceback=(
                ''.join(traceback.format_tb(exc.__traceback__))
                if getattr(exc, '__traceback__', None) is not None else None
            ),

            exc=None if omit_exc_object else exc,

            cmd=cmd,
        )


class CommandOutputOrException(Abstract, ta.Generic[CommandOutputT]):
    @property
    @abc.abstractmethod
    def output(self) -> ta.Optional[CommandOutputT]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def exception(self) -> ta.Optional[CommandException]:
        raise NotImplementedError


@dc.dataclass(frozen=True)
class CommandOutputOrExceptionData(CommandOutputOrException):
    output: ta.Optional[Command.Output] = None
    exception: ta.Optional[CommandException] = None


class CommandExecutor(Abstract, ta.Generic[CommandT, CommandOutputT]):
    @abc.abstractmethod
    def execute(self, cmd: CommandT) -> ta.Awaitable[CommandOutputT]:
        raise NotImplementedError

    async def try_execute(
            self,
            cmd: CommandT,
            *,
            log: ta.Optional[logging.Logger] = None,
            omit_exc_object: bool = False,
    ) -> CommandOutputOrException[CommandOutputT]:
        try:
            o = await self.execute(cmd)

        except Exception as e:  # noqa
            if log is not None:
                log.exception('Exception executing command: %r', type(cmd))

            return CommandOutputOrExceptionData(exception=CommandException.of(
                e,
                omit_exc_object=omit_exc_object,
                cmd=cmd,
            ))

        else:
            return CommandOutputOrExceptionData(output=o)


##


@dc.dataclass(frozen=True)
class CommandRegistration:
    command_cls: ta.Type[Command]

    name: ta.Optional[str] = None

    @property
    def name_or_default(self) -> str:
        if not (cls_name := self.command_cls.__name__).endswith('Command'):
            raise NameError(cls_name)
        return snake_case(cls_name[:-len('Command')])


CommandRegistrations = ta.NewType('CommandRegistrations', ta.Sequence[CommandRegistration])


##


@dc.dataclass(frozen=True)
class CommandExecutorRegistration:
    command_cls: ta.Type[Command]
    executor_cls: ta.Type[CommandExecutor]


CommandExecutorRegistrations = ta.NewType('CommandExecutorRegistrations', ta.Sequence[CommandExecutorRegistration])


##


CommandNameMap = ta.NewType('CommandNameMap', ta.Mapping[str, ta.Type[Command]])


def build_command_name_map(crs: CommandRegistrations) -> CommandNameMap:
    dct: ta.Dict[str, ta.Type[Command]] = {}
    cr: CommandRegistration
    for cr in crs:
        if (name := cr.name_or_default) in dct:
            raise NameError(name)
        dct[name] = cr.command_cls
    return CommandNameMap(dct)



# ruff: noqa: UP007 UP045
"""
TODO:
 - retries
"""
import json
import queue
import time
import typing as ta
import urllib.request

from omlish.lite.logs import log

from ....journald.messages import JournalctlMessage  # noqa
from ....threadworkers import ThreadWorker
from ..logs import AwsLogMessageBuilder
from ..logs import AwsPutLogEventsResponse
from .cursor import JournalctlToAwsCursor


##


class JournalctlToAwsPosterWorker(ThreadWorker):
    def __init__(
            self,
            queue,  # type: queue.Queue[ta.Sequence[JournalctlMessage]]  # noqa
            builder: AwsLogMessageBuilder,
            cursor: JournalctlToAwsCursor,
            *,
            ensure_locked: ta.Optional[ta.Callable[[], None]] = None,
            dry_run: bool = False,
            queue_timeout_s: float = 1.,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(**kwargs)

        self._queue = queue
        self._builder = builder
        self._cursor = cursor
        self._ensure_locked = ensure_locked
        self._dry_run = dry_run
        self._queue_timeout_s = queue_timeout_s
    #

    def _run(self) -> None:
        if self._ensure_locked is not None:
            self._ensure_locked()

        last_cursor: ta.Optional[str] = None  # noqa
        while True:
            self._heartbeat()

            try:
                msgs: ta.Sequence[JournalctlMessage] = self._queue.get(timeout=self._queue_timeout_s)
            except queue.Empty:
                msgs = []

            if not msgs:
                log.debug('Empty queue chunk')
                continue

            log.debug('%r', msgs)

            cur_cursor: ta.Optional[str] = None
            for m in reversed(msgs):
                if m.cursor is not None:
                    cur_cursor = m.cursor
                    break

            feed_msgs = []
            for m in msgs:
                feed_msgs.append(AwsLogMessageBuilder.Message(
                    message=json.dumps(m.dct, sort_keys=True),
                    ts_ms=int((m.ts_us / 1000.) if m.ts_us is not None else (time.time() * 1000.)),
                ))

            for post in self._builder.feed(feed_msgs):
                log.debug('%r', post)

                if not self._dry_run:
                    with urllib.request.urlopen(urllib.request.Request(  # noqa
                            post.url,
                            method='POST',
                            headers=dict(post.headers),
                            data=post.data,
                    )) as resp:
                        response = AwsPutLogEventsResponse.from_aws(json.loads(resp.read().decode('utf-8')))
                    log.debug('%r', response)

            if cur_cursor is not None:
                self._cursor.set(cur_cursor)
                last_cursor = cur_cursor  # noqa



# flake8: noqa: E501
# ruff: noqa: N801 S105
# fmt: off
# @omlish-generated
import enum as _enum  # noqa
import typing as _ta  # noqa

from omlish import dataclasses as _dc  # noqa

from .. import base as _base  # noqa


##


AcceptRanges = _ta.NewType('AcceptRanges', str)

AccountId = _ta.NewType('AccountId', str)

Body = _ta.NewType('Body', bytes)

BucketKeyEnabled = _ta.NewType('BucketKeyEnabled', bool)

BucketName = _ta.NewType('BucketName', str)

BucketRegion = _ta.NewType('BucketRegion', str)

BypassGovernanceRetention = _ta.NewType('BypassGovernanceRetention', bool)

CacheControl = _ta.NewType('CacheControl', str)


class ChecksumAlgorithm(_enum.Enum):
    CRC32 = 'CRC32'
    CRC32C = 'CRC32C'
    SHA1 = 'SHA1'
    SHA256 = 'SHA256'
    CRC64NVME = 'CRC64NVME'


ChecksumCRC32 = _ta.NewType('ChecksumCRC32', str)

ChecksumCRC32C = _ta.NewType('ChecksumCRC32C', str)

ChecksumCRC64NVME = _ta.NewType('ChecksumCRC64NVME', str)


class ChecksumMode(_enum.Enum):
    ENABLED = 'ENABLED'


ChecksumSHA1 = _ta.NewType('ChecksumSHA1', str)

ChecksumSHA256 = _ta.NewType('ChecksumSHA256', str)


class ChecksumType(_enum.Enum):
    COMPOSITE = 'COMPOSITE'
    FULL_OBJECT = 'FULL_OBJECT'


ContentDisposition = _ta.NewType('ContentDisposition', str)

ContentEncoding = _ta.NewType('ContentEncoding', str)

ContentLanguage = _ta.NewType('ContentLanguage', str)

ContentLength = _ta.NewType('ContentLength', int)

ContentMD5 = _ta.NewType('ContentMD5', str)

ContentRange = _ta.NewType('ContentRange', str)

ContentType = _ta.NewType('ContentType', str)

CreationDate = _ta.NewType('CreationDate', _base.Timestamp)

DeleteMarker = _ta.NewType('DeleteMarker', bool)

Delimiter = _ta.NewType('Delimiter', str)

DisplayName = _ta.NewType('DisplayName', str)

ETag = _ta.NewType('ETag', str)


class EncodingType(_enum.Enum):
    URL = 'url'


@_dc.dataclass(frozen=True, kw_only=True)
class EncryptionTypeMismatch(
    _base.Shape,
    shape_name='EncryptionTypeMismatch',
):
    pass


Expiration = _ta.NewType('Expiration', str)

Expires = _ta.NewType('Expires', _base.Timestamp)

FetchOwner = _ta.NewType('FetchOwner', bool)

GrantFullControl = _ta.NewType('GrantFullControl', str)

GrantRead = _ta.NewType('GrantRead', str)

GrantReadACP = _ta.NewType('GrantReadACP', str)

GrantWriteACP = _ta.NewType('GrantWriteACP', str)

ID = _ta.NewType('ID', str)

IfMatch = _ta.NewType('IfMatch', str)

IfMatchLastModifiedTime = _ta.NewType('IfMatchLastModifiedTime', _base.Timestamp)

IfMatchSize = _ta.NewType('IfMatchSize', int)

IfModifiedSince = _ta.NewType('IfModifiedSince', _base.Timestamp)

IfNoneMatch = _ta.NewType('IfNoneMatch', str)

IfUnmodifiedSince = _ta.NewType('IfUnmodifiedSince', _base.Timestamp)


class IntelligentTieringAccessTier(_enum.Enum):
    ARCHIVE_ACCESS = 'ARCHIVE_ACCESS'
    DEEP_ARCHIVE_ACCESS = 'DEEP_ARCHIVE_ACCESS'


@_dc.dataclass(frozen=True, kw_only=True)
class InvalidRequest(
    _base.Shape,
    shape_name='InvalidRequest',
):
    pass


@_dc.dataclass(frozen=True, kw_only=True)
class InvalidWriteOffset(
    _base.Shape,
    shape_name='InvalidWriteOffset',
):
    pass


IsRestoreInProgress = _ta.NewType('IsRestoreInProgress', bool)

IsTruncated = _ta.NewType('IsTruncated', bool)

KeyCount = _ta.NewType('KeyCount', int)

LastModified = _ta.NewType('LastModified', _base.Timestamp)

MFA = _ta.NewType('MFA', str)

MaxBuckets = _ta.NewType('MaxBuckets', int)

MaxKeys = _ta.NewType('MaxKeys', int)

MetadataKey = _ta.NewType('MetadataKey', str)

MetadataValue = _ta.NewType('MetadataValue', str)

MissingMeta = _ta.NewType('MissingMeta', int)

NextToken = _ta.NewType('NextToken', str)


@_dc.dataclass(frozen=True, kw_only=True)
class NoSuchBucket(
    _base.Shape,
    shape_name='NoSuchBucket',
):
    pass


@_dc.dataclass(frozen=True, kw_only=True)
class NoSuchKey(
    _base.Shape,
    shape_name='NoSuchKey',
):
    pass


class ObjectCannedACL(_enum.Enum):
    PRIVATE = 'private'
    PUBLIC_READ = 'public-read'
    PUBLIC_READ_WRITE = 'public-read-write'
    AUTHENTICATED_READ = 'authenticated-read'
    AWS_EXEC_READ = 'aws-exec-read'
    BUCKET_OWNER_READ = 'bucket-owner-read'
    BUCKET_OWNER_FULL_CONTROL = 'bucket-owner-full-control'


ObjectKey = _ta.NewType('ObjectKey', str)


class ObjectLockLegalHoldStatus(_enum.Enum):
    ON = 'ON'
    OFF = 'OFF'


class ObjectLockMode(_enum.Enum):
    GOVERNANCE = 'GOVERNANCE'
    COMPLIANCE = 'COMPLIANCE'


ObjectLockRetainUntilDate = _ta.NewType('ObjectLockRetainUntilDate', _base.Timestamp)


class ObjectStorageClass(_enum.Enum):
    STANDARD = 'STANDARD'
    REDUCED_REDUNDANCY = 'REDUCED_REDUNDANCY'
    GLACIER = 'GLACIER'
    STANDARD_IA = 'STANDARD_IA'
    ONEZONE_IA = 'ONEZONE_IA'
    INTELLIGENT_TIERING = 'INTELLIGENT_TIERING'
    DEEP_ARCHIVE = 'DEEP_ARCHIVE'
    OUTPOSTS = 'OUTPOSTS'
    GLACIER_IR = 'GLACIER_IR'
    SNOW = 'SNOW'
    EXPRESS_ONEZONE = 'EXPRESS_ONEZONE'
    FSX_OPENZFS = 'FSX_OPENZFS'


ObjectVersionId = _ta.NewType('ObjectVersionId', str)


class OptionalObjectAttributes(_enum.Enum):
    RESTORE_STATUS = 'RestoreStatus'


PartNumber = _ta.NewType('PartNumber', int)

PartsCount = _ta.NewType('PartsCount', int)

Prefix = _ta.NewType('Prefix', str)

Range = _ta.NewType('Range', str)


class ReplicationStatus(_enum.Enum):
    COMPLETE = 'COMPLETE'
    PENDING = 'PENDING'
    FAILED = 'FAILED'
    REPLICA = 'REPLICA'
    COMPLETED = 'COMPLETED'


class RequestCharged(_enum.Enum):
    REQUESTER = 'requester'


class RequestPayer(_enum.Enum):
    REQUESTER = 'requester'


ResponseCacheControl = _ta.NewType('ResponseCacheControl', str)

ResponseContentDisposition = _ta.NewType('ResponseContentDisposition', str)

ResponseContentEncoding = _ta.NewType('ResponseContentEncoding', str)

ResponseContentLanguage = _ta.NewType('ResponseContentLanguage', str)

ResponseContentType = _ta.NewType('ResponseContentType', str)

ResponseExpires = _ta.NewType('ResponseExpires', _base.Timestamp)

Restore = _ta.NewType('Restore', str)

RestoreExpiryDate = _ta.NewType('RestoreExpiryDate', _base.Timestamp)

S3RegionalOrS3ExpressBucketArnString = _ta.NewType('S3RegionalOrS3ExpressBucketArnString', str)

SSECustomerAlgorithm = _ta.NewType('SSECustomerAlgorithm', str)

SSECustomerKey = _ta.NewType('SSECustomerKey', str)

SSECustomerKeyMD5 = _ta.NewType('SSECustomerKeyMD5', str)

SSEKMSEncryptionContext = _ta.NewType('SSEKMSEncryptionContext', str)

SSEKMSKeyId = _ta.NewType('SSEKMSKeyId', str)


class ServerSideEncryption(_enum.Enum):
    AES256 = 'AES256'
    AWS_FSX = 'aws:fsx'
    AWS_KMS = 'aws:kms'
    AWS_KMS_DSSE = 'aws:kms:dsse'


Size = _ta.NewType('Size', int)

StartAfter = _ta.NewType('StartAfter', str)


class StorageClass(_enum.Enum):
    STANDARD = 'STANDARD'
    REDUCED_REDUNDANCY = 'REDUCED_REDUNDANCY'
    STANDARD_IA = 'STANDARD_IA'
    ONEZONE_IA = 'ONEZONE_IA'
    INTELLIGENT_TIERING = 'INTELLIGENT_TIERING'
    GLACIER = 'GLACIER'
    DEEP_ARCHIVE = 'DEEP_ARCHIVE'
    OUTPOSTS = 'OUTPOSTS'
    GLACIER_IR = 'GLACIER_IR'
    SNOW = 'SNOW'
    EXPRESS_ONEZONE = 'EXPRESS_ONEZONE'
    FSX_OPENZFS = 'FSX_OPENZFS'


TagCount = _ta.NewType('TagCount', int)

TaggingHeader = _ta.NewType('TaggingHeader', str)

Token = _ta.NewType('Token', str)


@_dc.dataclass(frozen=True, kw_only=True)
class TooManyParts(
    _base.Shape,
    shape_name='TooManyParts',
):
    pass


WebsiteRedirectLocation = _ta.NewType('WebsiteRedirectLocation', str)

WriteOffsetBytes = _ta.NewType('WriteOffsetBytes', int)


@_dc.dataclass(frozen=True, kw_only=True)
class Bucket(
    _base.Shape,
    shape_name='Bucket',
):
    name: BucketName | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Name',
        shape_name='BucketName',
    ))

    creation_date: CreationDate | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='CreationDate',
        shape_name='CreationDate',
    ))

    bucket_region: BucketRegion | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='BucketRegion',
        shape_name='BucketRegion',
    ))

    bucket_arn: S3RegionalOrS3ExpressBucketArnString | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='BucketArn',
        shape_name='S3RegionalOrS3ExpressBucketArnString',
    ))


ChecksumAlgorithmList: _ta.TypeAlias = _ta.Sequence[ChecksumAlgorithm]


@_dc.dataclass(frozen=True, kw_only=True)
class CommonPrefix(
    _base.Shape,
    shape_name='CommonPrefix',
):
    prefix: Prefix | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Prefix',
        shape_name='Prefix',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class DeleteObjectOutput(
    _base.Shape,
    shape_name='DeleteObjectOutput',
):
    delete_marker: DeleteMarker | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='DeleteMarker',
        serialization_name='x-amz-delete-marker',
        shape_name='DeleteMarker',
    ))

    version_id: ObjectVersionId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='VersionId',
        serialization_name='x-amz-version-id',
        shape_name='ObjectVersionId',
    ))

    request_charged: RequestCharged | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RequestCharged',
        serialization_name='x-amz-request-charged',
        shape_name='RequestCharged',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class DeleteObjectRequest(
    _base.Shape,
    shape_name='DeleteObjectRequest',
):
    bucket: BucketName = _dc.field(metadata=_base.field_metadata(
        member_name='Bucket',
        serialization_name='Bucket',
        shape_name='BucketName',
    ))

    key: ObjectKey = _dc.field(metadata=_base.field_metadata(
        member_name='Key',
        serialization_name='Key',
        shape_name='ObjectKey',
    ))

    mfa: MFA | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='MFA',
        serialization_name='x-amz-mfa',
        shape_name='MFA',
    ))

    version_id: ObjectVersionId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='VersionId',
        serialization_name='versionId',
        shape_name='ObjectVersionId',
    ))

    request_payer: RequestPayer | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RequestPayer',
        serialization_name='x-amz-request-payer',
        shape_name='RequestPayer',
    ))

    bypass_governance_retention: BypassGovernanceRetention | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='BypassGovernanceRetention',
        serialization_name='x-amz-bypass-governance-retention',
        shape_name='BypassGovernanceRetention',
    ))

    expected_bucket_owner: AccountId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ExpectedBucketOwner',
        serialization_name='x-amz-expected-bucket-owner',
        shape_name='AccountId',
    ))

    if_match: IfMatch | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IfMatch',
        serialization_name='If-Match',
        shape_name='IfMatch',
    ))

    if_match_last_modified_time: IfMatchLastModifiedTime | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IfMatchLastModifiedTime',
        serialization_name='x-amz-if-match-last-modified-time',
        shape_name='IfMatchLastModifiedTime',
    ))

    if_match_size: IfMatchSize | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IfMatchSize',
        serialization_name='x-amz-if-match-size',
        shape_name='IfMatchSize',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class GetObjectRequest(
    _base.Shape,
    shape_name='GetObjectRequest',
):
    bucket: BucketName = _dc.field(metadata=_base.field_metadata(
        member_name='Bucket',
        serialization_name='Bucket',
        shape_name='BucketName',
    ))

    if_match: IfMatch | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IfMatch',
        serialization_name='If-Match',
        shape_name='IfMatch',
    ))

    if_modified_since: IfModifiedSince | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IfModifiedSince',
        serialization_name='If-Modified-Since',
        shape_name='IfModifiedSince',
    ))

    if_none_match: IfNoneMatch | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IfNoneMatch',
        serialization_name='If-None-Match',
        shape_name='IfNoneMatch',
    ))

    if_unmodified_since: IfUnmodifiedSince | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IfUnmodifiedSince',
        serialization_name='If-Unmodified-Since',
        shape_name='IfUnmodifiedSince',
    ))

    key: ObjectKey = _dc.field(metadata=_base.field_metadata(
        member_name='Key',
        serialization_name='Key',
        shape_name='ObjectKey',
    ))

    range: Range | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Range',
        serialization_name='Range',
        shape_name='Range',
    ))

    response_cache_control: ResponseCacheControl | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ResponseCacheControl',
        serialization_name='response-cache-control',
        shape_name='ResponseCacheControl',
    ))

    response_content_disposition: ResponseContentDisposition | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ResponseContentDisposition',
        serialization_name='response-content-disposition',
        shape_name='ResponseContentDisposition',
    ))

    response_content_encoding: ResponseContentEncoding | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ResponseContentEncoding',
        serialization_name='response-content-encoding',
        shape_name='ResponseContentEncoding',
    ))

    response_content_language: ResponseContentLanguage | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ResponseContentLanguage',
        serialization_name='response-content-language',
        shape_name='ResponseContentLanguage',
    ))

    response_content_type: ResponseContentType | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ResponseContentType',
        serialization_name='response-content-type',
        shape_name='ResponseContentType',
    ))

    response_expires: ResponseExpires | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ResponseExpires',
        serialization_name='response-expires',
        shape_name='ResponseExpires',
    ))

    version_id: ObjectVersionId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='VersionId',
        serialization_name='versionId',
        shape_name='ObjectVersionId',
    ))

    sse_customer_algorithm: SSECustomerAlgorithm | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerAlgorithm',
        serialization_name='x-amz-server-side-encryption-customer-algorithm',
        shape_name='SSECustomerAlgorithm',
    ))

    sse_customer_key: SSECustomerKey | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerKey',
        serialization_name='x-amz-server-side-encryption-customer-key',
        shape_name='SSECustomerKey',
    ))

    sse_customer_key_md5: SSECustomerKeyMD5 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerKeyMD5',
        serialization_name='x-amz-server-side-encryption-customer-key-MD5',
        shape_name='SSECustomerKeyMD5',
    ))

    request_payer: RequestPayer | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RequestPayer',
        serialization_name='x-amz-request-payer',
        shape_name='RequestPayer',
    ))

    part_number: PartNumber | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='PartNumber',
        serialization_name='partNumber',
        shape_name='PartNumber',
    ))

    expected_bucket_owner: AccountId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ExpectedBucketOwner',
        serialization_name='x-amz-expected-bucket-owner',
        shape_name='AccountId',
    ))

    checksum_mode: ChecksumMode | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumMode',
        serialization_name='x-amz-checksum-mode',
        shape_name='ChecksumMode',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class InvalidObjectState(
    _base.Shape,
    shape_name='InvalidObjectState',
):
    storage_class: StorageClass | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='StorageClass',
        shape_name='StorageClass',
    ))

    access_tier: IntelligentTieringAccessTier | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='AccessTier',
        shape_name='IntelligentTieringAccessTier',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class ListBucketsRequest(
    _base.Shape,
    shape_name='ListBucketsRequest',
):
    max_buckets: MaxBuckets | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='MaxBuckets',
        serialization_name='max-buckets',
        shape_name='MaxBuckets',
    ))

    continuation_token: Token | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContinuationToken',
        serialization_name='continuation-token',
        shape_name='Token',
    ))

    prefix: Prefix | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Prefix',
        serialization_name='prefix',
        shape_name='Prefix',
    ))

    bucket_region: BucketRegion | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='BucketRegion',
        serialization_name='bucket-region',
        shape_name='BucketRegion',
    ))


Metadata: _ta.TypeAlias = _ta.Mapping[MetadataKey, MetadataKey]

OptionalObjectAttributesList: _ta.TypeAlias = _ta.Sequence[OptionalObjectAttributes]


@_dc.dataclass(frozen=True, kw_only=True)
class Owner(
    _base.Shape,
    shape_name='Owner',
):
    display_name: DisplayName | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='DisplayName',
        shape_name='DisplayName',
    ))

    i_d: ID | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ID',
        shape_name='ID',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class PutObjectOutput(
    _base.Shape,
    shape_name='PutObjectOutput',
):
    expiration: Expiration | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Expiration',
        serialization_name='x-amz-expiration',
        shape_name='Expiration',
    ))

    etag: ETag | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ETag',
        serialization_name='ETag',
        shape_name='ETag',
    ))

    checksum_crc32: ChecksumCRC32 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumCRC32',
        serialization_name='x-amz-checksum-crc32',
        shape_name='ChecksumCRC32',
    ))

    checksum_crc32c: ChecksumCRC32C | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumCRC32C',
        serialization_name='x-amz-checksum-crc32c',
        shape_name='ChecksumCRC32C',
    ))

    checksum_crc64_nvme: ChecksumCRC64NVME | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumCRC64NVME',
        serialization_name='x-amz-checksum-crc64nvme',
        shape_name='ChecksumCRC64NVME',
    ))

    checksum_sha1: ChecksumSHA1 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumSHA1',
        serialization_name='x-amz-checksum-sha1',
        shape_name='ChecksumSHA1',
    ))

    checksum_sha256: ChecksumSHA256 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumSHA256',
        serialization_name='x-amz-checksum-sha256',
        shape_name='ChecksumSHA256',
    ))

    checksum_type: ChecksumType | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumType',
        serialization_name='x-amz-checksum-type',
        shape_name='ChecksumType',
    ))

    server_side_encryption: ServerSideEncryption | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ServerSideEncryption',
        serialization_name='x-amz-server-side-encryption',
        shape_name='ServerSideEncryption',
    ))

    version_id: ObjectVersionId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='VersionId',
        serialization_name='x-amz-version-id',
        shape_name='ObjectVersionId',
    ))

    sse_customer_algorithm: SSECustomerAlgorithm | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerAlgorithm',
        serialization_name='x-amz-server-side-encryption-customer-algorithm',
        shape_name='SSECustomerAlgorithm',
    ))

    sse_customer_key_md5: SSECustomerKeyMD5 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerKeyMD5',
        serialization_name='x-amz-server-side-encryption-customer-key-MD5',
        shape_name='SSECustomerKeyMD5',
    ))

    sse_kms_key_id: SSEKMSKeyId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSEKMSKeyId',
        serialization_name='x-amz-server-side-encryption-aws-kms-key-id',
        shape_name='SSEKMSKeyId',
    ))

    sse_kms_encryption_context: SSEKMSEncryptionContext | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSEKMSEncryptionContext',
        serialization_name='x-amz-server-side-encryption-context',
        shape_name='SSEKMSEncryptionContext',
    ))

    bucket_key_enabled: BucketKeyEnabled | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='BucketKeyEnabled',
        serialization_name='x-amz-server-side-encryption-bucket-key-enabled',
        shape_name='BucketKeyEnabled',
    ))

    size: Size | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Size',
        serialization_name='x-amz-object-size',
        shape_name='Size',
    ))

    request_charged: RequestCharged | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RequestCharged',
        serialization_name='x-amz-request-charged',
        shape_name='RequestCharged',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class RestoreStatus(
    _base.Shape,
    shape_name='RestoreStatus',
):
    is_restore_in_progress: IsRestoreInProgress | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IsRestoreInProgress',
        shape_name='IsRestoreInProgress',
    ))

    restore_expiry_date: RestoreExpiryDate | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RestoreExpiryDate',
        shape_name='RestoreExpiryDate',
    ))


Buckets: _ta.TypeAlias = _ta.Sequence[Bucket]

CommonPrefixList: _ta.TypeAlias = _ta.Sequence[CommonPrefix]


@_dc.dataclass(frozen=True, kw_only=True)
class GetObjectOutput(
    _base.Shape,
    shape_name='GetObjectOutput',
):
    body: Body | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Body',
        shape_name='Body',
    ))

    delete_marker: DeleteMarker | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='DeleteMarker',
        serialization_name='x-amz-delete-marker',
        shape_name='DeleteMarker',
    ))

    accept_ranges: AcceptRanges | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='AcceptRanges',
        serialization_name='accept-ranges',
        shape_name='AcceptRanges',
    ))

    expiration: Expiration | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Expiration',
        serialization_name='x-amz-expiration',
        shape_name='Expiration',
    ))

    restore: Restore | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Restore',
        serialization_name='x-amz-restore',
        shape_name='Restore',
    ))

    last_modified: LastModified | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='LastModified',
        serialization_name='Last-Modified',
        shape_name='LastModified',
    ))

    content_length: ContentLength | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentLength',
        serialization_name='Content-Length',
        shape_name='ContentLength',
    ))

    etag: ETag | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ETag',
        serialization_name='ETag',
        shape_name='ETag',
    ))

    checksum_crc32: ChecksumCRC32 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumCRC32',
        serialization_name='x-amz-checksum-crc32',
        shape_name='ChecksumCRC32',
    ))

    checksum_crc32c: ChecksumCRC32C | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumCRC32C',
        serialization_name='x-amz-checksum-crc32c',
        shape_name='ChecksumCRC32C',
    ))

    checksum_crc64_nvme: ChecksumCRC64NVME | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumCRC64NVME',
        serialization_name='x-amz-checksum-crc64nvme',
        shape_name='ChecksumCRC64NVME',
    ))

    checksum_sha1: ChecksumSHA1 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumSHA1',
        serialization_name='x-amz-checksum-sha1',
        shape_name='ChecksumSHA1',
    ))

    checksum_sha256: ChecksumSHA256 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumSHA256',
        serialization_name='x-amz-checksum-sha256',
        shape_name='ChecksumSHA256',
    ))

    checksum_type: ChecksumType | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumType',
        serialization_name='x-amz-checksum-type',
        shape_name='ChecksumType',
    ))

    missing_meta: MissingMeta | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='MissingMeta',
        serialization_name='x-amz-missing-meta',
        shape_name='MissingMeta',
    ))

    version_id: ObjectVersionId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='VersionId',
        serialization_name='x-amz-version-id',
        shape_name='ObjectVersionId',
    ))

    cache_control: CacheControl | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='CacheControl',
        serialization_name='Cache-Control',
        shape_name='CacheControl',
    ))

    content_disposition: ContentDisposition | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentDisposition',
        serialization_name='Content-Disposition',
        shape_name='ContentDisposition',
    ))

    content_encoding: ContentEncoding | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentEncoding',
        serialization_name='Content-Encoding',
        shape_name='ContentEncoding',
    ))

    content_language: ContentLanguage | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentLanguage',
        serialization_name='Content-Language',
        shape_name='ContentLanguage',
    ))

    content_range: ContentRange | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentRange',
        serialization_name='Content-Range',
        shape_name='ContentRange',
    ))

    content_type: ContentType | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentType',
        serialization_name='Content-Type',
        shape_name='ContentType',
    ))

    expires: Expires | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Expires',
        serialization_name='Expires',
        shape_name='Expires',
    ))

    website_redirect_location: WebsiteRedirectLocation | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='WebsiteRedirectLocation',
        serialization_name='x-amz-website-redirect-location',
        shape_name='WebsiteRedirectLocation',
    ))

    server_side_encryption: ServerSideEncryption | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ServerSideEncryption',
        serialization_name='x-amz-server-side-encryption',
        shape_name='ServerSideEncryption',
    ))

    metadata: Metadata | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Metadata',
        serialization_name='x-amz-meta-',
        value_type=_base.MapValueType(MetadataKey, MetadataValue),
        shape_name='Metadata',
    ))

    sse_customer_algorithm: SSECustomerAlgorithm | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerAlgorithm',
        serialization_name='x-amz-server-side-encryption-customer-algorithm',
        shape_name='SSECustomerAlgorithm',
    ))

    sse_customer_key_md5: SSECustomerKeyMD5 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerKeyMD5',
        serialization_name='x-amz-server-side-encryption-customer-key-MD5',
        shape_name='SSECustomerKeyMD5',
    ))

    sse_kms_key_id: SSEKMSKeyId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSEKMSKeyId',
        serialization_name='x-amz-server-side-encryption-aws-kms-key-id',
        shape_name='SSEKMSKeyId',
    ))

    bucket_key_enabled: BucketKeyEnabled | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='BucketKeyEnabled',
        serialization_name='x-amz-server-side-encryption-bucket-key-enabled',
        shape_name='BucketKeyEnabled',
    ))

    storage_class: StorageClass | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='StorageClass',
        serialization_name='x-amz-storage-class',
        shape_name='StorageClass',
    ))

    request_charged: RequestCharged | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RequestCharged',
        serialization_name='x-amz-request-charged',
        shape_name='RequestCharged',
    ))

    replication_status: ReplicationStatus | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ReplicationStatus',
        serialization_name='x-amz-replication-status',
        shape_name='ReplicationStatus',
    ))

    parts_count: PartsCount | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='PartsCount',
        serialization_name='x-amz-mp-parts-count',
        shape_name='PartsCount',
    ))

    tag_count: TagCount | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='TagCount',
        serialization_name='x-amz-tagging-count',
        shape_name='TagCount',
    ))

    object_lock_mode: ObjectLockMode | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ObjectLockMode',
        serialization_name='x-amz-object-lock-mode',
        shape_name='ObjectLockMode',
    ))

    object_lock_retain_until_date: ObjectLockRetainUntilDate | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ObjectLockRetainUntilDate',
        serialization_name='x-amz-object-lock-retain-until-date',
        shape_name='ObjectLockRetainUntilDate',
    ))

    object_lock_legal_hold_status: ObjectLockLegalHoldStatus | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ObjectLockLegalHoldStatus',
        serialization_name='x-amz-object-lock-legal-hold',
        shape_name='ObjectLockLegalHoldStatus',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class ListObjectsV2Request(
    _base.Shape,
    shape_name='ListObjectsV2Request',
):
    bucket: BucketName = _dc.field(metadata=_base.field_metadata(
        member_name='Bucket',
        serialization_name='Bucket',
        shape_name='BucketName',
    ))

    delimiter: Delimiter | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Delimiter',
        serialization_name='delimiter',
        shape_name='Delimiter',
    ))

    encoding_type: EncodingType | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='EncodingType',
        serialization_name='encoding-type',
        shape_name='EncodingType',
    ))

    max_keys: MaxKeys | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='MaxKeys',
        serialization_name='max-keys',
        shape_name='MaxKeys',
    ))

    prefix: Prefix | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Prefix',
        serialization_name='prefix',
        shape_name='Prefix',
    ))

    continuation_token: Token | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContinuationToken',
        serialization_name='continuation-token',
        shape_name='Token',
    ))

    fetch_owner: FetchOwner | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='FetchOwner',
        serialization_name='fetch-owner',
        shape_name='FetchOwner',
    ))

    start_after: StartAfter | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='StartAfter',
        serialization_name='start-after',
        shape_name='StartAfter',
    ))

    request_payer: RequestPayer | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RequestPayer',
        serialization_name='x-amz-request-payer',
        shape_name='RequestPayer',
    ))

    expected_bucket_owner: AccountId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ExpectedBucketOwner',
        serialization_name='x-amz-expected-bucket-owner',
        shape_name='AccountId',
    ))

    optional_object_attributes: OptionalObjectAttributesList | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='OptionalObjectAttributes',
        serialization_name='x-amz-optional-object-attributes',
        value_type=_base.ListValueType(OptionalObjectAttributes),
        shape_name='OptionalObjectAttributesList',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class Object(
    _base.Shape,
    shape_name='Object',
):
    key: ObjectKey | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Key',
        shape_name='ObjectKey',
    ))

    last_modified: LastModified | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='LastModified',
        shape_name='LastModified',
    ))

    etag: ETag | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ETag',
        shape_name='ETag',
    ))

    checksum_algorithm: ChecksumAlgorithmList | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumAlgorithm',
        value_type=_base.ListValueType(ChecksumAlgorithm),
        shape_name='ChecksumAlgorithmList',
    ))

    checksum_type: ChecksumType | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumType',
        shape_name='ChecksumType',
    ))

    size: Size | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Size',
        shape_name='Size',
    ))

    storage_class: ObjectStorageClass | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='StorageClass',
        shape_name='ObjectStorageClass',
    ))

    owner: Owner | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Owner',
        shape_name='Owner',
    ))

    restore_status: RestoreStatus | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RestoreStatus',
        shape_name='RestoreStatus',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class PutObjectRequest(
    _base.Shape,
    shape_name='PutObjectRequest',
):
    acl: ObjectCannedACL | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ACL',
        serialization_name='x-amz-acl',
        shape_name='ObjectCannedACL',
    ))

    body: Body | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Body',
        shape_name='Body',
    ))

    bucket: BucketName = _dc.field(metadata=_base.field_metadata(
        member_name='Bucket',
        serialization_name='Bucket',
        shape_name='BucketName',
    ))

    cache_control: CacheControl | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='CacheControl',
        serialization_name='Cache-Control',
        shape_name='CacheControl',
    ))

    content_disposition: ContentDisposition | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentDisposition',
        serialization_name='Content-Disposition',
        shape_name='ContentDisposition',
    ))

    content_encoding: ContentEncoding | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentEncoding',
        serialization_name='Content-Encoding',
        shape_name='ContentEncoding',
    ))

    content_language: ContentLanguage | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentLanguage',
        serialization_name='Content-Language',
        shape_name='ContentLanguage',
    ))

    content_length: ContentLength | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentLength',
        serialization_name='Content-Length',
        shape_name='ContentLength',
    ))

    content_md5: ContentMD5 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentMD5',
        serialization_name='Content-MD5',
        shape_name='ContentMD5',
    ))

    content_type: ContentType | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContentType',
        serialization_name='Content-Type',
        shape_name='ContentType',
    ))

    checksum_algorithm: ChecksumAlgorithm | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumAlgorithm',
        serialization_name='x-amz-sdk-checksum-algorithm',
        shape_name='ChecksumAlgorithm',
    ))

    checksum_crc32: ChecksumCRC32 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumCRC32',
        serialization_name='x-amz-checksum-crc32',
        shape_name='ChecksumCRC32',
    ))

    checksum_crc32c: ChecksumCRC32C | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumCRC32C',
        serialization_name='x-amz-checksum-crc32c',
        shape_name='ChecksumCRC32C',
    ))

    checksum_crc64_nvme: ChecksumCRC64NVME | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumCRC64NVME',
        serialization_name='x-amz-checksum-crc64nvme',
        shape_name='ChecksumCRC64NVME',
    ))

    checksum_sha1: ChecksumSHA1 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumSHA1',
        serialization_name='x-amz-checksum-sha1',
        shape_name='ChecksumSHA1',
    ))

    checksum_sha256: ChecksumSHA256 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ChecksumSHA256',
        serialization_name='x-amz-checksum-sha256',
        shape_name='ChecksumSHA256',
    ))

    expires: Expires | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Expires',
        serialization_name='Expires',
        shape_name='Expires',
    ))

    if_match: IfMatch | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IfMatch',
        serialization_name='If-Match',
        shape_name='IfMatch',
    ))

    if_none_match: IfNoneMatch | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IfNoneMatch',
        serialization_name='If-None-Match',
        shape_name='IfNoneMatch',
    ))

    grant_full_control: GrantFullControl | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='GrantFullControl',
        serialization_name='x-amz-grant-full-control',
        shape_name='GrantFullControl',
    ))

    grant_read: GrantRead | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='GrantRead',
        serialization_name='x-amz-grant-read',
        shape_name='GrantRead',
    ))

    grant_read_acp: GrantReadACP | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='GrantReadACP',
        serialization_name='x-amz-grant-read-acp',
        shape_name='GrantReadACP',
    ))

    grant_write_acp: GrantWriteACP | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='GrantWriteACP',
        serialization_name='x-amz-grant-write-acp',
        shape_name='GrantWriteACP',
    ))

    key: ObjectKey = _dc.field(metadata=_base.field_metadata(
        member_name='Key',
        serialization_name='Key',
        shape_name='ObjectKey',
    ))

    write_offset_bytes: WriteOffsetBytes | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='WriteOffsetBytes',
        serialization_name='x-amz-write-offset-bytes',
        shape_name='WriteOffsetBytes',
    ))

    metadata: Metadata | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Metadata',
        serialization_name='x-amz-meta-',
        value_type=_base.MapValueType(MetadataKey, MetadataValue),
        shape_name='Metadata',
    ))

    server_side_encryption: ServerSideEncryption | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ServerSideEncryption',
        serialization_name='x-amz-server-side-encryption',
        shape_name='ServerSideEncryption',
    ))

    storage_class: StorageClass | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='StorageClass',
        serialization_name='x-amz-storage-class',
        shape_name='StorageClass',
    ))

    website_redirect_location: WebsiteRedirectLocation | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='WebsiteRedirectLocation',
        serialization_name='x-amz-website-redirect-location',
        shape_name='WebsiteRedirectLocation',
    ))

    sse_customer_algorithm: SSECustomerAlgorithm | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerAlgorithm',
        serialization_name='x-amz-server-side-encryption-customer-algorithm',
        shape_name='SSECustomerAlgorithm',
    ))

    sse_customer_key: SSECustomerKey | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerKey',
        serialization_name='x-amz-server-side-encryption-customer-key',
        shape_name='SSECustomerKey',
    ))

    sse_customer_key_md5: SSECustomerKeyMD5 | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSECustomerKeyMD5',
        serialization_name='x-amz-server-side-encryption-customer-key-MD5',
        shape_name='SSECustomerKeyMD5',
    ))

    sse_kms_key_id: SSEKMSKeyId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSEKMSKeyId',
        serialization_name='x-amz-server-side-encryption-aws-kms-key-id',
        shape_name='SSEKMSKeyId',
    ))

    sse_kms_encryption_context: SSEKMSEncryptionContext | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='SSEKMSEncryptionContext',
        serialization_name='x-amz-server-side-encryption-context',
        shape_name='SSEKMSEncryptionContext',
    ))

    bucket_key_enabled: BucketKeyEnabled | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='BucketKeyEnabled',
        serialization_name='x-amz-server-side-encryption-bucket-key-enabled',
        shape_name='BucketKeyEnabled',
    ))

    request_payer: RequestPayer | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RequestPayer',
        serialization_name='x-amz-request-payer',
        shape_name='RequestPayer',
    ))

    tagging: TaggingHeader | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Tagging',
        serialization_name='x-amz-tagging',
        shape_name='TaggingHeader',
    ))

    object_lock_mode: ObjectLockMode | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ObjectLockMode',
        serialization_name='x-amz-object-lock-mode',
        shape_name='ObjectLockMode',
    ))

    object_lock_retain_until_date: ObjectLockRetainUntilDate | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ObjectLockRetainUntilDate',
        serialization_name='x-amz-object-lock-retain-until-date',
        shape_name='ObjectLockRetainUntilDate',
    ))

    object_lock_legal_hold_status: ObjectLockLegalHoldStatus | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ObjectLockLegalHoldStatus',
        serialization_name='x-amz-object-lock-legal-hold',
        shape_name='ObjectLockLegalHoldStatus',
    ))

    expected_bucket_owner: AccountId | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ExpectedBucketOwner',
        serialization_name='x-amz-expected-bucket-owner',
        shape_name='AccountId',
    ))


@_dc.dataclass(frozen=True, kw_only=True)
class ListBucketsOutput(
    _base.Shape,
    shape_name='ListBucketsOutput',
):
    buckets: Buckets | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Buckets',
        value_type=_base.ListValueType(Bucket),
        shape_name='Buckets',
    ))

    owner: Owner | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Owner',
        shape_name='Owner',
    ))

    continuation_token: NextToken | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContinuationToken',
        shape_name='NextToken',
    ))

    prefix: Prefix | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Prefix',
        shape_name='Prefix',
    ))


ObjectList: _ta.TypeAlias = _ta.Sequence[Object]


@_dc.dataclass(frozen=True, kw_only=True)
class ListObjectsV2Output(
    _base.Shape,
    shape_name='ListObjectsV2Output',
):
    is_truncated: IsTruncated | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='IsTruncated',
        shape_name='IsTruncated',
    ))

    contents: ObjectList | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Contents',
        value_type=_base.ListValueType(Object),
        shape_name='ObjectList',
    ))

    name: BucketName | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Name',
        shape_name='BucketName',
    ))

    prefix: Prefix | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Prefix',
        shape_name='Prefix',
    ))

    delimiter: Delimiter | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='Delimiter',
        shape_name='Delimiter',
    ))

    max_keys: MaxKeys | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='MaxKeys',
        shape_name='MaxKeys',
    ))

    common_prefixes: CommonPrefixList | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='CommonPrefixes',
        value_type=_base.ListValueType(CommonPrefix),
        shape_name='CommonPrefixList',
    ))

    encoding_type: EncodingType | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='EncodingType',
        shape_name='EncodingType',
    ))

    key_count: KeyCount | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='KeyCount',
        shape_name='KeyCount',
    ))

    continuation_token: Token | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='ContinuationToken',
        shape_name='Token',
    ))

    next_continuation_token: NextToken | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='NextContinuationToken',
        shape_name='NextToken',
    ))

    start_after: StartAfter | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='StartAfter',
        shape_name='StartAfter',
    ))

    request_charged: RequestCharged | None = _dc.field(default=None, metadata=_base.field_metadata(
        member_name='RequestCharged',
        serialization_name='x-amz-request-charged',
        shape_name='RequestCharged',
    ))


ALL_SHAPES: frozenset[type[_base.Shape]] = frozenset([
    Bucket,
    CommonPrefix,
    DeleteObjectOutput,
    DeleteObjectRequest,
    EncryptionTypeMismatch,
    GetObjectOutput,
    GetObjectRequest,
    InvalidObjectState,
    InvalidRequest,
    InvalidWriteOffset,
    ListBucketsOutput,
    ListBucketsRequest,
    ListObjectsV2Output,
    ListObjectsV2Request,
    NoSuchBucket,
    NoSuchKey,
    Object,
    Owner,
    PutObjectOutput,
    PutObjectRequest,
    RestoreStatus,
    TooManyParts,
])


##


DELETE_OBJECT = _base.Operation(
    name='DeleteObject',
    input=DeleteObjectRequest,
    output=DeleteObjectOutput,
)

GET_OBJECT = _base.Operation(
    name='GetObject',
    input=GetObjectRequest,
    output=GetObjectOutput,
    errors=[
        InvalidObjectState,
        NoSuchKey,
    ],
)

LIST_BUCKETS = _base.Operation(
    name='ListBuckets',
    input=ListBucketsRequest,
    output=ListBucketsOutput,
)

LIST_OBJECTS_V2 = _base.Operation(
    name='ListObjectsV2',
    input=ListObjectsV2Request,
    output=ListObjectsV2Output,
    errors=[
        NoSuchBucket,
    ],
)

PUT_OBJECT = _base.Operation(
    name='PutObject',
    input=PutObjectRequest,
    output=PutObjectOutput,
    errors=[
        EncryptionTypeMismatch,
        InvalidRequest,
        InvalidWriteOffset,
        TooManyParts,
    ],
)


ALL_OPERATIONS: frozenset[_base.Operation] = frozenset([
    DELETE_OBJECT,
    GET_OBJECT,
    LIST_BUCKETS,
    LIST_OBJECTS_V2,
    PUT_OBJECT,
])



# ruff: noqa: UP006 UP007 UP045
import abc
import asyncio
import json
import struct
import typing as ta

from omlish.lite.abstract import Abstract
from omlish.lite.json import json_dumps_compact
from omlish.lite.marshal import OBJ_MARSHALER_MANAGER
from omlish.lite.marshal import ObjMarshalerManager


T = ta.TypeVar('T')


##


class RemoteChannel(Abstract):
    @abc.abstractmethod
    def send_obj(self, o: ta.Any, ty: ta.Any = None) -> ta.Awaitable[None]:
        raise NotImplementedError

    @abc.abstractmethod
    def recv_obj(self, ty: ta.Type[T]) -> ta.Awaitable[ta.Optional[T]]:
        raise NotImplementedError

    def set_marshaler(self, msh: ObjMarshalerManager) -> None:  # noqa
        pass


##


class RemoteChannelImpl(RemoteChannel):
    def __init__(
            self,
            input: asyncio.StreamReader,  # noqa
            output: asyncio.StreamWriter,
            *,
            msh: ObjMarshalerManager = OBJ_MARSHALER_MANAGER,
    ) -> None:
        super().__init__()

        self._input = input
        self._output = output
        self._msh = msh

        self._input_lock = asyncio.Lock()
        self._output_lock = asyncio.Lock()

    def set_marshaler(self, msh: ObjMarshalerManager) -> None:
        self._msh = msh

    #

    async def _send_obj(self, o: ta.Any, ty: ta.Any = None) -> None:
        j = json_dumps_compact(self._msh.marshal_obj(o, ty))
        d = j.encode('utf-8')

        self._output.write(struct.pack('<I', len(d)))
        self._output.write(d)
        await self._output.drain()

    async def send_obj(self, o: ta.Any, ty: ta.Any = None) -> None:
        async with self._output_lock:
            return await self._send_obj(o, ty)

    #

    async def _recv_obj(self, ty: ta.Type[T]) -> ta.Optional[T]:
        d = await self._input.read(4)
        if not d:
            return None
        if len(d) != 4:
            raise EOFError

        sz = struct.unpack('<I', d)[0]
        d = await self._input.read(sz)
        if len(d) != sz:
            raise EOFError

        j = json.loads(d.decode('utf-8'))
        return self._msh.unmarshal_obj(j, ty)

    async def recv_obj(self, ty: ta.Type[T]) -> ta.Optional[T]:
        async with self._input_lock:
            return await self._recv_obj(ty)



# ruff: noqa: UP006 UP007 UP045
import asyncio
import functools
import logging
import os
import signal
import threading
import time
import typing as ta

from omlish.asyncs.asyncio.streams import asyncio_open_stream_reader
from omlish.asyncs.asyncio.streams import asyncio_open_stream_writer
from omlish.lite.cached import cached_nullary
from omlish.lite.check import check
from omlish.lite.inject import Injector
from omlish.lite.logs import log
from omlish.lite.marshal import ObjMarshalerManager
from omlish.lite.pycharm import pycharm_debug_connect
from omlish.os.deathsig import set_process_deathsig

from ...pyremote import pyremote_bootstrap_finalize
from ..bootstrap import MainBootstrap
from ..commands.local import LocalCommandExecutor
from .channel import RemoteChannel
from .channel import RemoteChannelImpl
from .execution import _RemoteCommandHandler
from .execution import _RemoteLogHandler


if ta.TYPE_CHECKING:
    from ..bootstrap_ import main_bootstrap
else:
    main_bootstrap: ta.Any = None


##


class _RemoteExecutionLogHandler(logging.Handler):
    def __init__(self, fn: ta.Callable[[str], None]) -> None:
        super().__init__()

        self._fn = fn

    def emit(self, record):
        msg = self.format(record)
        self._fn(msg)


##


class _RemoteExecutionMain:
    def __init__(
            self,
            chan: RemoteChannel,
    ) -> None:
        super().__init__()

        self._chan = chan

        self.__bootstrap: ta.Optional[MainBootstrap] = None
        self.__injector: ta.Optional[Injector] = None

    @property
    def _bootstrap(self) -> MainBootstrap:
        return check.not_none(self.__bootstrap)

    @property
    def _injector(self) -> Injector:
        return check.not_none(self.__injector)

    #

    def _timebomb_main(
            self,
            delay_s: float,
            *,
            sig: int = signal.SIGINT,
            code: int = 1,
    ) -> None:
        time.sleep(delay_s)

        if (pgid := os.getpgid(0)) == os.getpid():
            os.killpg(pgid, sig)

        os._exit(code)  # noqa

    @cached_nullary
    def _timebomb_thread(self) -> ta.Optional[threading.Thread]:
        if (tbd := self._bootstrap.remote_config.timebomb_delay_s) is None:
            return None

        thr = threading.Thread(
            target=functools.partial(self._timebomb_main, tbd),
            name=f'{self.__class__.__name__}.timebomb',
            daemon=True,
        )

        thr.start()

        log.debug('Started timebomb thread: %r', thr)

        return thr

    #

    @cached_nullary
    def _log_handler(self) -> _RemoteLogHandler:
        return _RemoteLogHandler(self._chan)

    #

    async def _setup(self) -> None:
        check.none(self.__bootstrap)
        check.none(self.__injector)

        # Bootstrap

        self.__bootstrap = check.not_none(await self._chan.recv_obj(MainBootstrap))

        if (prd := self._bootstrap.remote_config.pycharm_remote_debug) is not None:
            pycharm_debug_connect(prd)

        self.__injector = main_bootstrap(self._bootstrap)

        self._chan.set_marshaler(self._injector[ObjMarshalerManager])

        # Post-bootstrap

        if self._bootstrap.remote_config.set_pgid:
            if os.getpgid(0) != os.getpid():
                log.debug('Setting pgid')
                os.setpgid(0, 0)

        if (ds := self._bootstrap.remote_config.deathsig) is not None:
            log.debug('Setting deathsig: %s', ds)
            set_process_deathsig(int(signal.Signals[f'SIG{ds.upper()}']))

        self._timebomb_thread()

        if self._bootstrap.remote_config.forward_logging:
            log.debug('Installing log forwarder')
            logging.root.addHandler(self._log_handler())

    #

    async def run(self) -> None:
        await self._setup()

        executor = self._injector[LocalCommandExecutor]

        handler = _RemoteCommandHandler(self._chan, executor)

        await handler.run()


def _remote_execution_main() -> None:
    rt = pyremote_bootstrap_finalize()  # noqa

    async def inner() -> None:
        input = await asyncio_open_stream_reader(rt.input)  # noqa
        output = await asyncio_open_stream_writer(rt.output)

        chan = RemoteChannelImpl(
            input,
            output,
        )

        await _RemoteExecutionMain(chan).run()

    asyncio.run(inner())



# ruff: noqa: UP006 UP007 UP045
import dataclasses as dc
import fcntl
import os
import typing as ta

from .utils.fds import close_fd
from .utils.fds import make_pipe
from .utils.ostypes import Fd


##


@dc.dataclass(frozen=True)
class ProcessPipes:
    child_stdin: ta.Optional[Fd] = None
    stdin: ta.Optional[Fd] = None

    stdout: ta.Optional[Fd] = None
    child_stdout: ta.Optional[Fd] = None

    stderr: ta.Optional[Fd] = None
    child_stderr: ta.Optional[Fd] = None

    def child_fds(self) -> ta.List[Fd]:
        return [fd for fd in [self.child_stdin, self.child_stdout, self.child_stderr] if fd is not None]

    def parent_fds(self) -> ta.List[Fd]:
        return [fd for fd in [self.stdin, self.stdout, self.stderr] if fd is not None]


def make_process_pipes(stderr=True) -> ProcessPipes:
    """
    Create pipes for parent to child stdin/stdout/stderr communications. Open fd in non-blocking mode so we can read
    them in the mainloop without blocking. If stderr is False, don't create a pipe for stderr.
    """

    pipes: ta.Dict[str, ta.Optional[Fd]] = {
        'child_stdin': None,
        'stdin': None,

        'stdout': None,
        'child_stdout': None,

        'stderr': None,
        'child_stderr': None,
    }

    try:
        pipes['child_stdin'], pipes['stdin'] = make_pipe()
        pipes['stdout'], pipes['child_stdout'] = make_pipe()

        if stderr:
            pipes['stderr'], pipes['child_stderr'] = make_pipe()

        for fd in (
                pipes['stdout'],
                pipes['stderr'],
                pipes['stdin'],
        ):
            if fd is not None:
                flags = fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NDELAY
                fcntl.fcntl(fd, fcntl.F_SETFL, flags)

        return ProcessPipes(**pipes)

    except OSError:
        for fd in pipes.values():
            if fd is not None:
                close_fd(fd)

        raise


def close_pipes(pipes: ProcessPipes) -> None:
    close_parent_pipes(pipes)
    close_child_pipes(pipes)


def close_parent_pipes(pipes: ProcessPipes) -> None:
    for fd in pipes.parent_fds():
        close_fd(fd)


def close_child_pipes(pipes: ProcessPipes) -> None:
    for fd in pipes.child_fds():
        close_fd(fd)



# ruff: noqa: UP006 UP007 UP043 UP045
import asyncio
import contextlib
import typing as ta

from omlish.asyncs.asyncio.channels import asyncio_create_bytes_channel
from omlish.lite.cached import cached_nullary
from omlish.lite.contextmanagers import aclosing
from omlish.lite.marshal import ObjMarshalerManager

from ...pyremote import PyremoteBootstrapDriver
from ...pyremote import PyremoteBootstrapOptions
from ...pyremote import pyremote_build_bootstrap_cmd
from ..bootstrap import MainBootstrap
from ..commands.local import LocalCommandExecutor
from ._main import _remote_execution_main  # noqa
from .channel import RemoteChannelImpl
from .execution import RemoteCommandExecutor
from .execution import _RemoteCommandHandler
from .payload import RemoteExecutionPayloadFile
from .payload import get_remote_payload_src
from .spawning import RemoteSpawning


##


class PyremoteRemoteExecutionConnector:
    def __init__(
            self,
            *,
            spawning: RemoteSpawning,
            msh: ObjMarshalerManager,
            payload_file: ta.Optional[RemoteExecutionPayloadFile] = None,
    ) -> None:
        super().__init__()

        self._spawning = spawning
        self._msh = msh
        self._payload_file = payload_file

    #

    @cached_nullary
    def _payload_src(self) -> str:
        return get_remote_payload_src(file=self._payload_file)

    @cached_nullary
    def _remote_src(self) -> ta.Sequence[str]:
        return [
            self._payload_src(),
            '_remote_execution_main()',
        ]

    @cached_nullary
    def _spawn_src(self) -> str:
        return pyremote_build_bootstrap_cmd(__package__ or 'manage')

    #

    @contextlib.asynccontextmanager
    async def connect(
            self,
            tgt: RemoteSpawning.Target,
            bs: MainBootstrap,
    ) -> ta.AsyncGenerator[RemoteCommandExecutor, None]:
        spawn_src = self._spawn_src()
        remote_src = self._remote_src()

        async with self._spawning.spawn(
                tgt,
                spawn_src,
                debug=bs.main_config.debug,
        ) as proc:
            res = await PyremoteBootstrapDriver(  # noqa
                remote_src,
                PyremoteBootstrapOptions(
                    debug=bs.main_config.debug,
                ),
            ).async_run(
                proc.stdout,
                proc.stdin,
            )

            chan = RemoteChannelImpl(
                proc.stdout,
                proc.stdin,
                msh=self._msh,
            )

            await chan.send_obj(bs)

            rce: RemoteCommandExecutor
            async with aclosing(RemoteCommandExecutor(chan)) as rce:
                await rce.start()

                yield rce


##


class InProcessRemoteExecutionConnector:
    def __init__(
            self,
            *,
            msh: ObjMarshalerManager,
            local_executor: LocalCommandExecutor,
    ) -> None:
        super().__init__()

        self._msh = msh
        self._local_executor = local_executor

    @contextlib.asynccontextmanager
    async def connect(self) -> ta.AsyncGenerator[RemoteCommandExecutor, None]:
        r0, w0 = asyncio_create_bytes_channel()
        r1, w1 = asyncio_create_bytes_channel()

        remote_chan = RemoteChannelImpl(r0, w1, msh=self._msh)
        local_chan = RemoteChannelImpl(r1, w0, msh=self._msh)

        rch = _RemoteCommandHandler(
            remote_chan,
            self._local_executor,
        )
        rch_task = asyncio.create_task(rch.run())  # noqa
        try:
            rce: RemoteCommandExecutor
            async with aclosing(RemoteCommandExecutor(local_chan)) as rce:
                await rce.start()

                yield rce

        finally:
            rch.stop()
            await rch_task



# ruff: noqa: UP006 UP007 UP045
import signal

from omlish.lite.logs import log

from .groups import ProcessGroupManager
from .states import SupervisorState
from .types import ProcessOutputDispatcher
from .types import SupervisorStateManager
from .utils.signals import SignalReceiver
from .utils.signals import sig_name


##


class SignalHandler:
    def __init__(
            self,
            *,
            states: SupervisorStateManager,
            signal_receiver: SignalReceiver,
            process_groups: ProcessGroupManager,
    ) -> None:
        super().__init__()

        self._states = states
        self._signal_receiver = signal_receiver
        self._process_groups = process_groups

    def set_signals(self) -> None:
        self._signal_receiver.install(
            signal.SIGTERM,
            signal.SIGINT,
            signal.SIGQUIT,
            signal.SIGHUP,
            signal.SIGCHLD,
            signal.SIGUSR2,
        )

    def handle_signals(self) -> None:
        sig = self._signal_receiver.get_signal()
        if not sig:
            return

        if sig in (signal.SIGTERM, signal.SIGINT, signal.SIGQUIT):
            log.warning('received %s indicating exit request', sig_name(sig))
            self._states.set_state(SupervisorState.SHUTDOWN)

        elif sig == signal.SIGHUP:
            if self._states.state == SupervisorState.SHUTDOWN:
                log.warning('ignored %s indicating restart request (shutdown in progress)', sig_name(sig))  # noqa
            else:
                log.warning('received %s indicating restart request', sig_name(sig))  # noqa
                self._states.set_state(SupervisorState.RESTARTING)

        elif sig == signal.SIGCHLD:
            log.debug('received %s indicating a child quit', sig_name(sig))

        elif sig == signal.SIGUSR2:
            log.info('received %s indicating log reopen request', sig_name(sig))

            for p in self._process_groups.all_processes():
                for d in p.get_dispatchers():
                    if isinstance(d, ProcessOutputDispatcher):
                        d.reopen_logs()

        else:
            log.debug('received %s indicating nothing', sig_name(sig))
